/*
 * physmem.c
 *
 * 	Christophe Lucas <clucas@rotomalug.org>
 * 	This file is under GPLv2 terms.
 */

#include <testos/list.h>
#include <testos/klibc.h>
#include <testos/physmem.h>
#include <testos/errno.h>
#include <testos/types.h>
#include <video/x86_video.h>
#include <video/bochs.h>

#define KERNEL_CODE_BASE	((paddr_t)MEM_ALIGN_INF(&__b_kernel))
#define KERNEL_CODE_END		((paddr_t)MEM_ALIGN_SUP(&__e_kernel))

extern char __b_kernel, __e_kernel;

static struct physmem_page* free_pages_list;
static ui32_t nr_total_pages;
static struct physmem_page* used_pages_list;
static ui32_t nr_used_pages;
static struct physmem_page* tab_physmem_desc_page;

static paddr_t paddr_base_mem, paddr_top_mem;
static ui32_t ram_size = 0;
static paddr_t kernel_code_base = 0;
static paddr_t kernel_code_end = 0;

ui32_t get_nr_free_pages()
{
	return nr_total_pages - nr_used_pages;
}

ui32_t get_nr_used_pages()
{
	return nr_used_pages;
}

paddr_t get_physmem_page()
{
	struct physmem_page *p;
	if (get_nr_free_pages() == 0 ) {
		bochs_printf("physmem: %s() : Il n'y a plus de mmoire.\n",__FUNCTION__);
		return (paddr_t)NULL;
	} else {
		p = list_pop_head(free_pages_list);
		list_add_tail(used_pages_list, p);
		nr_used_pages++;
		p->refcounter++;
		return p->paddr;
	}
}

/*
 * It gives us one desc_page by means of his index in tab_physmem_desc_page
 */
static struct physmem_page* get_physmem_page_at(ui32_t at)
{
	if (at > (ram_size>>PAGE_SHIFT)) {
		printk("physmem: %s : param at is invalid or out of bound\n", __FUNCTION__);
		return NULL;
	} else {
		return tab_physmem_desc_page + at ;
	}
}

/*
 * Ok we keep one desc_page by means of his address
 */
static struct physmem_page* get_physmem_page_at_address(paddr_t addr)
{
	if (addr > ram_size) {
		printk("physmem: %s : 'addr' is out of %d ram_size.\n", 
					__FUNCTION__, (int)ram_size);
		return NULL;
	} else {
		return get_physmem_page_at(addr >> PAGE_SHIFT);
	}
}

int release_physmem_page(paddr_t paddr)
{
	struct physmem_page *page = get_physmem_page_at_address(paddr);

	if (paddr > ram_size) {
		printk("physmem: %s : pointer to page to release is NULL", __FUNCTION__);
		return -EINVAL;
	} else {
		page->refcounter--;
		if (!page->refcounter) {
			list_delete(used_pages_list, page);
			nr_used_pages--;
			list_add_tail(free_pages_list, page);
		}
		return 0;
	}
	return -EINVAL;	
}


static paddr_t get_paddr_kernel_code_base()
{
	return kernel_code_base;
}

static paddr_t get_paddr_kernel_code_end()
{
	return kernel_code_end;
}

int physmem_init(ui32_t param_ram_size)
{
	enum {
		PHYSMEM_RESERVED,
		PHYSMEM_HWMAP,
		PHYSMEM_KERNEL,
		PHYSMEM_FREE,
	} page_type;
	struct physmem_page *work = NULL;
	ui32_t i = 0;

	ram_size = param_ram_size;
	paddr_base_mem = MEM_ALIGN_INF(PAGE_SIZE);
	paddr_top_mem = MEM_ALIGN_INF(ram_size);
	nr_total_pages = nr_used_pages = 0;
	kernel_code_base = KERNEL_CODE_BASE;
	kernel_code_end = KERNEL_CODE_END + 
			MEM_ALIGN_SUP(sizeof(struct physmem_page) * (ram_size>>PAGE_SHIFT));


	list_init(free_pages_list);
	list_init(used_pages_list);

	printk("physmem: physical memory subsystem initialized.\n");
	bochs_printf("physmem: paddr_base_mem=%x - paddr_top_mem=%x - nr frame mem unit=%d\n",
				paddr_base_mem, paddr_top_mem, ALIGN_SUP(ram_size,PAGE_SIZE));


	/* Little problem with kernel code end limit */
	if (kernel_code_end > paddr_top_mem) {
		return -ENOMEM;
	}

	tab_physmem_desc_page = (struct physmem_page*)KERNEL_CODE_END;

	for(i=0, work = tab_physmem_desc_page;
		i<paddr_top_mem;
		i += PAGE_SIZE, work++) {

		memcpy(work, 0x0, sizeof(*work));
		work->refcounter = 0;

		page_type = PHYSMEM_FREE;

		/* reserved area 0 ... physical base mem */
		if (i<paddr_base_mem) {
			page_type = PHYSMEM_RESERVED;
		}

		if ((i >= MEM_BIOS_AND_VIDEO_START) &&
			(i<MEM_BIOS_AND_VIDEO_END)) {
			page_type = PHYSMEM_HWMAP;
		}

		if ((i>=(int)kernel_code_base) && (i<(int)kernel_code_end)) {
			page_type = PHYSMEM_KERNEL;
		}

		switch(page_type) {
			case PHYSMEM_KERNEL:
			case PHYSMEM_RESERVED:
			case PHYSMEM_HWMAP:
				nr_used_pages++;
				work->paddr = (paddr_t) i;
				(work->refcounter)++;
				list_add_tail(used_pages_list, work);
				break;
			case PHYSMEM_FREE:
				work->paddr = (paddr_t) i;
				list_add_tail(free_pages_list, work);
				memcpy((void*)work->paddr, 0x0, PAGE_SHIFT);
				break;
			default: printk("physmem: problem with page type :-(\n");
		}
		nr_total_pages++;
	}

	printk("physmem: free pages=%d - used pages=%d\n",
				(int)get_nr_free_pages(), (int)nr_used_pages);
	printk("physmem: paddr_base_mem=0x%x - paddr_top_mem=0x%x\n",paddr_base_mem, paddr_top_mem);
	printk("physmem: __b_kernel=0x%x - __e_kernel+sizeof(tab_physmem_desc_page)=0x%x\n", kernel_code_base, kernel_code_end);
	printk("physmem: tab_physmem_desc_page=0x%x\n", (paddr_t)tab_physmem_desc_page);

	return OK;
}

/*
 * 511*4*2 == 4088
 * 4088 + 4 + 4 == 4096
 */ 
#define TEST_STRUCT_SIZE 511
struct test_struct {
	ui32_t before[TEST_STRUCT_SIZE];
	struct test_struct *prev, *next;
	ui32_t after[TEST_STRUCT_SIZE];
};

void test_physmem_subsystem()
{
	int i;
	struct physmem_page *page;
	struct test_struct *test_page;
	struct test_struct *list;

	page = get_physmem_page_at( (ram_size >> PAGE_SHIFT) + 100 );

	page = get_physmem_page_at(500);

	bochs_printf("physmem: %s free pages=%d - used pages=%d\n", __FUNCTION__,
				(int)get_nr_free_pages(), (int)nr_used_pages);

	list = NULL;
	while ((test_page = (struct test_struct*)get_physmem_page()) != NULL) {
		for (i=0;i<TEST_STRUCT_SIZE;i++) {
			test_page->before[i] = test_page->after[i] = (ui32_t)test_page;
		}
		list_add_tail(list, test_page);
	}
	bochs_printf("physmem: free pages=%d - used pages=%d\n",
				(int)get_nr_free_pages(), (int)nr_used_pages);

	while( (test_page = (struct test_struct*)list_pop_head(list)) != NULL ) {
		for (i=0;i<TEST_STRUCT_SIZE;i++) {
			if ( (test_page->after[i] != (ui32_t)test_page) || 
					(test_page->before[i] != (ui32_t)test_page)) {
				bochs_printf("physmem: %s : page(0x%x) was overwritten.\n",
						__FUNCTION__, (paddr_t)test_page);
				return ;
			}
		}
		release_physmem_page((paddr_t)test_page);
	}

	bochs_printf("physmem: Aprs avoir remis librer toutes les pages.\n");
	bochs_printf("physmem: free pages=%d - used pages=%d\n", 
				(int)get_nr_free_pages(), (int)nr_used_pages);

	memcpy((void*)page->paddr, "500AA\0", 6);

	release_physmem_page((paddr_t)page);
	bochs_printf("physmem: free pages=%d - used pages=%d\n", 
				(int)get_nr_free_pages(), (int)nr_used_pages);
}

